@@ -1,11 +1,11 @@
 ;;; julia-repl.el --- A minor mode for a Julia REPL -*- lexical-binding:t; no-byte-compile:t -*-
 
-;; Copyright (C) 2016  Tamas K. Papp
+;; Copyright (C) 2016–2024 Tamas K. Papp
 ;; Author: Tamas Papp <tkpapp@gmail.com>
 ;; Keywords: languages
-;; Package-Version: 1.3.0
-;; Package-Revision: 7ce38a9caf2a
-;; Package-Requires: ((emacs "25.1")(s "1.12"))
+;; Package-Version: 1.5.2
+;; Package-Revision: 0173237a43d9
+;; Package-Requires: ((emacs "29.1")(s "1.12"))
 ;; URL: https://github.com/tpapp/julia-repl
 
 ;;; Usage:
@@ -48,13 +48,12 @@
 (require 'seq)
 (require 'subr-x)
 
-
 ;;
 ;; customizations
 ;;
 
 (defgroup julia-repl nil
-  "A minor mode for a Julia REPL"
+  "A minor mode for a Julia REPL."
   :group 'julia)
 
 (defcustom julia-repl-inferior-buffer-name-base "julia"
@@ -65,7 +64,7 @@ See ‘julia-repl--inferior-buffer-name’."
   :group 'julia-repl)
 
 (defcustom julia-repl-hook nil
-  "Hook to run after starting a Julia REPL term buffer."
+  "Hook to run after starting a Julia REPL term buffer. For all functions, it can be assumed that the current buffer is an inferior Julia buffer with a running REPL."
   :type 'hook
   :group 'julia-repl)
 
@@ -86,12 +85,20 @@ Note that this affects all buffers using the ‘ansi-term’ map."
   :type 'boolean
   :group 'julia-repl)
 
+(defcustom julia-repl-pop-to-buffer t
+  "When non-nil, pop to `julia-repl'."
+  :type 'symbol
+  :group 'julia-repl)
+
 (defcustom julia-repl-path-rewrite-rules nil
   "A list of rewrite rules applied to paths sent with ‘include’ and similar.
 
-Each rule should be a function that is called with a path, in the given order. If the function returns ‘nil’, the next one is tried, otherwise the result is used as the rewritten path.
+Each rule should be a function that is called with a path, in the given
+order. If the function returns nil, the next one is tried, otherwise
+the result is used as the rewritten path.
 
-This can be used a workaround, usually necessary on non-Unix systems. Eg for Cygwin-based Windows, use
+This can be used a workaround, usually necessary on non-Unix systems. Eg
+for Cygwin-based Windows, use
 
    (setq julia-repl-path-rewrite-rules julia-repl-cygwin-path-rewrite-rules)
 
@@ -105,7 +112,12 @@ in your Emacs init file after loading this package."
   :group 'julia-repl)
 
 (defcustom julia-repl-set-term-escape t
-  "Set the escape char C-x globally for term. Useful for switching windows, but users who do not want this globally should set it to nil."
+  "Set the escape char \`C-x globally for term. Useful for switching windows, but users who do not want this globally should set it to nil."
+  :type 'boolean
+  :group 'julia-repl)
+
+(defcustom julia-repl-skip-comments nil
+  "Make some send commands (currently `julia-repl-send-line' and `julia-repl-send-region-or-line') skip comments."
   :type 'boolean
   :group 'julia-repl)
 
@@ -190,11 +202,6 @@ When PASTE-P, “bracketed paste” mode will be used. When RET-P, terminate wit
     (goto-char (point))
     (compilation-next-error-function n reset))
 
-  (defun julia-repl--has-running-vterm-process (inferior-buffer)
-    "Return non-nil if ‘inferior-buffer’ has a running vterm process."
-    (let ((proc (buffer-local-value 'vterm--process inferior-buffer)))
-      (and proc (memq (process-status proc) '(run stop open listen connect)))))
-
   (cl-defstruct julia-repl--buffer-vterm
     "Terminal backend using ‘vterm’, which needs to be installed and loaded.")
 
@@ -203,8 +210,7 @@ When PASTE-P, “bracketed paste” mode will be used. When RET-P, terminate wit
     (if-let ((inferior-buffer (get-buffer (julia-repl--add-earmuffs name))))
         (with-current-buffer inferior-buffer
           (cl-assert (eq major-mode 'vterm-mode) nil "Expected vterm-mode. Changed mode or backends?")
-          ;; cf https://github.com/akermu/emacs-libvterm/issues/270
-          (when (julia-repl--has-running-vterm-process inferior-buffer)
+          (when (vterm-check-proc inferior-buffer)
             inferior-buffer))))
 
   (cl-defmethod julia-repl--make-buffer ((_terminal-backend julia-repl--buffer-vterm)
@@ -214,6 +220,7 @@ When PASTE-P, “bracketed paste” mode will be used. When RET-P, terminate wit
       (with-current-buffer vterm-buffer
         (let ((vterm-shell (s-join " " (cons executable-path switches))))
           (vterm-mode)
+          (local-set-key (kbd "C-c C-z") #'julia-repl--switch-back)
           ;; NOTE workaround for https://github.com/akermu/emacs-libvterm/issues/316, remove when fixed
           (add-hook 'compilation-shell-minor-mode-hook
                     ;; NOTE run *after* vterm's hook and overwrite `next-error-function'
@@ -228,22 +235,89 @@ When PASTE-P, “bracketed paste” mode will be used. When RET-P, terminate wit
       (when ret-p
         (vterm-send-return)))))
 
+;;; eat term
+
+(with-eval-after-load 'eat
+
+  (cl-defstruct julia-repl--buffer-eat
+    "Terminal backend using ‘eat’, which needs to be installed and loaded.")
+
+  (cl-defmethod julia-repl--locate-live-buffer ((_terminal-backend julia-repl--buffer-eat)
+						name)
+    (if-let ((inferior-buffer (get-buffer (julia-repl--add-earmuffs name))))
+	(with-current-buffer inferior-buffer
+	  (cl-assert (eq major-mode 'eat-mode) nil "Expected eat-mode. Changed mode or backends?")
+	  (when (eat-term-live-p inferior-buffer)
+	    inferior-buffer))))
+
+  (cl-defmethod julia-repl--make-buffer ((_terminal-backend julia-repl--buffer-eat)
+					 name executable-path switches)
+    (let ((inferior-buffer (apply #'eat-make name executable-path nil switches)))
+      (with-current-buffer inferior-buffer
+	(mapc (lambda (k)
+		(define-key eat-semi-char-mode-map k (global-key-binding k)))
+	      julia-repl-captures)
+	(local-set-key (kbd "C-c C-z") #'julia-repl--switch-back))
+      inferior-buffer))
+
+  (cl-defmethod julia-repl--send-to-backend ((_terminal-backend julia-repl--buffer-eat)
+					     buffer string paste-p ret-p)
+    (with-current-buffer buffer
+      (if paste-p
+	  (eat-term-send-string-as-yank eat-terminal string)
+        (eat-term-send-string eat-terminal string))
+      (when ret-p
+	(eat-term-send-string eat-terminal "\^M")))))
+
+;;; compiler output regexps for navigation
+
+(defconst julia-repl--CR-path
+  (rx (+ (not (any space ">" "<" "(" ")" "\t" "\n" "," "'" "\"" ";" ":"))))
+  "A regexp matching paths (in Julia stacktraces).")
+
+(defconst julia-repl--CR-at
+  (rx "@" space
+      (? (group (one-or-more (or (any "._") alnum))) space) ; group 1: module name
+      (group (regexp julia-repl--CR-path)) ; group 2: path
+      ":"
+      (group (+ num))                   ; group 3: line number
+      )
+  "Matches “@ Foo ~/code/Foo/src/Foo.jl:100”. This is what is used in Julia >= 1.6")
+
+(defconst julia-repl--CR-expression-starting
+  (rx "in expression starting at"
+      (+ space)
+      (group (regexp julia-repl--CR-path)) ; group 1: path
+      ":"
+      (group (+ num))                   ; group 2: line
+      )
+  "Matches “in expression starting at ~/code/Foo/src/Foo.jl:100”.")
+
+(defconst julia-repl--CR-filename
+  (rx (one-or-more (not (any " ><()\t\n,'\";:"))))
+  "An attempt to match filenames in error, info,  and warning messages printed by Julia.")
+
+(defconst julia-repl--CR-load-error
+  (rx
+   "while loading "
+   (group (regexp julia-repl--CR-filename))
+   ", in expression starting on line "
+   (group (one-or-more digit)))
+  "Compilation regexp matching “while loading /tmp/Foo.jl, in expression starting on line 2”.")
+
+(defconst julia-repl--CR-around
+  (rx
+   (or "around" "at" "Revise")
+   " "
+   (group (regexp julia-repl--CR-filename))
+   ":"
+   (group (one-or-more digit)))
+  "Compilation regexp matching “around /tmp/Foo.jl:2”, also starting with “at or “Revise”")
+
 ;;
 ;; global variables
 ;;
 
-(defvar julia-repl--compilation-regexp-alist
-  '(;; matches "while loading /tmp/Foo.jl, in expression starting on line 2"
-    (julia-load-error . ("while loading \\([^ ><()\t\n,'\";:]+\\), in expression starting on line \\([0-9]+\\)" 1 2))
-    ;; matches "around /tmp/Foo.jl:2", also starting with "at" or "Revise"
-    (julia-loc . ("\\(around\\|at\\|Revise\\) \\([^ ><()\t\n,'\";:]+\\):\\([0-9]+\\)" 2 3))
-    ;; matches "omitting file /tmp/Foo.jl due to parsing error near line 2", from Revise.parse_source!
-    (julia-warn-revise . ("omitting file \\([^ ><()\t\n,'\";:]+\\) due to parsing error near line \\([0-9]+\\)" 1 2))
-    )
-  "Specifications for highlighting error locations.
-
-Uses function ‘compilation-shell-minor-mode’.")
-
 (defvar julia-repl--terminal-backend
   (make-julia-repl--buffer-ansi-term)
   "Terminal backend, for internal use. Set using `julia-repl-set-terminal-backend'.")
@@ -255,14 +329,20 @@ Valid backends are currently:
 
 - ‘ansi-term’, using the ANSI terminal built into Emacs.
 
-- ‘vterm’, which requires that vterm is installed. See URL ‘https://github.com/akermu/emacs-libvterm’."
+- ‘vterm’, which requires that vterm is installed. See URL ‘https://github.com/akermu/emacs-libvterm’.
+
+- ‘eat’, which requires that eat is installed. See URL ‘https://codeberg.org/akib/emacs-eat’."
   (interactive "S")
   (cl-case backend
-    ('ansi-term
+    (ansi-term
      (setq julia-repl--terminal-backend (make-julia-repl--buffer-ansi-term)))
-    ('vterm
+    (vterm
      (require 'vterm)
-     (setq julia-repl--terminal-backend (make-julia-repl--buffer-vterm)))
+     (setq julia-repl--terminal-backend (make-julia-repl--buffer-vterm))
+     (add-to-list 'vterm-eval-cmds '("julia-repl--show" julia-repl--show)))
+    (eat
+     (require 'eat)
+     (setq julia-repl--terminal-backend (make-julia-repl--buffer-eat)))
     (otherwise
      (error "Unrecognized backend “%s”." backend))))
 
@@ -274,7 +354,9 @@ Entries have the form
 
   (KEY EXECUTABLE-PATH :BASEDIR BASEDIR)
 
-A missing :BASEDIR will be completed automatically when first used.
+`executable-path' is invoked as is, so make sure that it is in the path available to Emacs.
+
+A missing :BASEDIR will be completed automatically when first used, and it assumed to be the same afterwards.
 
 This is used for key lookup for ‘julia-repl-executable-key’. The
 first entry is the default.")
@@ -298,6 +380,9 @@ used to generate a buffer name.")
 These are used for offering choices when selecting new suffix.
 For internal use only.")
 
+(defvar-local julia-repl--script-buffer nil
+  "Buffer active before calling `julia-repl'.")
+
 (defvar julia-repl-executable-key nil
   "Key for the executable associated with the buffer.
 
@@ -317,6 +402,21 @@ generate a buffer name.")
 Valid values are NIL or a string. These take effect the next time
 a new Julia process is started.")
 
+(defvar julia-repl-compilation-location-legacy nil
+  "Whether to include recognize various legacy error messages in compilation output.
+ Mainly useful if you are using Julia <1.6.")
+
+(defun julia-repl--compilation-regexp-alist ()
+  "Return an alist suitable for use in `compilation-error-regexp-alist' for recognizing Julia error locations.
+
+ Cf `julia-repl-compilation-location-legacy'."
+  (let ((regexp-alist `((,julia-repl--CR-at 2 3)
+                        (,julia-repl--CR-expression-starting 1 2))))
+    (if julia-repl-compilation-location-legacy
+        (cons regexp-alist
+              `((,julia-repl--CR-load-error 1 2) (,julia-repl--CR-around 1 2)))
+      regexp-alist)))
+
 ;;
 ;; REPL buffer creation and setup
 ;;
@@ -346,7 +446,8 @@ Note that ‘make-term’ surrounds this string by *'s when converted to a buffe
 (cl-defun julia-repl--capture-basedir (executable-path)
   "Attempt to obtain the Julia base directory by querying the Julia executable.
 
-When NIL, this was unsuccessful."
+When NIL, this was unsuccessful. In this case, debug information
+is printed to the *Messages* buffer."
   (let* ((prefix "OK") ; prefix is used to verify that there was no error and help with extraction
          (expr (concat "print(\"" prefix
                        "\" * normpath(joinpath(VERSION <= v\"0.7-\" ? JULIA_HOME : Sys.BINDIR, "
@@ -354,8 +455,19 @@ When NIL, this was unsuccessful."
          (switches " --history-file=no --startup-file=no -qe ")
          (maybe-basedir (shell-command-to-string
                          (concat executable-path switches (concat "'" expr "'")))))
-    (when (string-prefix-p prefix maybe-basedir)
-      (substring maybe-basedir (length prefix)))))
+    (if (string-prefix-p prefix maybe-basedir)
+        (substring maybe-basedir (length prefix))
+      (progn
+        (message "Julia basedir query returned “%s”" maybe-basedir)
+        nil))))
+
+(defun julia-repl--executable-path (executable-record)
+  "Retrun the Julia executable for the given EXECUTABLE-RECORD.
+
+It is checked for being a string."
+  (let ((executable-path (cl-second executable-record)))
+    (cl-assert (stringp executable-path) nil "No valid executable path found in %s" executable-record)
+    executable-path))
 
 (defun julia-repl--complete-executable-record! (executable-record)
   "Complete EXECUTABLE-RECORD if necessary.
@@ -365,9 +477,10 @@ Queries and appends missing information if necessary.
 Note: when cannot capture the base dir, it is set to NIL to
 prevent further attempts."
   (unless (plist-member (cddr executable-record) :basedir)
-    (let* ((executable-path (cl-second executable-record))
+    (let* ((executable-path (julia-repl--executable-path executable-record))
            (basedir (julia-repl--capture-basedir executable-path)))
-      (nconc executable-record `(:basedir ,basedir))
+      (when basedir
+        (nconc executable-record `(:basedir ,basedir)))
       (unless basedir
         (warn "could not capture basedir for Julia executable %s"
               executable-path)))))
@@ -377,10 +490,7 @@ prevent further attempts."
 
 BASEDIR is used for resolving relative paths."
   (with-current-buffer inferior-buffer
-    (setq-local compilation-error-regexp-alist-alist
-                julia-repl--compilation-regexp-alist)
-    (setq-local compilation-error-regexp-alist
-                (mapcar #'car compilation-error-regexp-alist-alist))
+    (setq-local compilation-error-regexp-alist (julia-repl--compilation-regexp-alist))
     (when basedir
       (setq-local compilation-search-path (list basedir)))
     (compilation-shell-minor-mode 1)))
@@ -521,7 +631,7 @@ Valid keys are the first items in ‘julia-repl-executable-records’."
       (let ((executable-record (julia-repl--executable-record executable-key))
             (switches julia-repl-switches))
         (julia-repl--complete-executable-record! executable-record)
-        (let* ((executable-path (cl-second executable-record))
+        (let* ((executable-path (julia-repl--executable-path executable-record))
                (basedir (plist-get (cddr executable-record) :basedir))
                (inferior-buffer (julia-repl--make-buffer terminal-backend name executable-path
                                                          (when switches
@@ -529,7 +639,8 @@ Valid keys are the first items in ‘julia-repl-executable-records’."
           (when julia-repl-compilation-mode
             (julia-repl--setup-compilation-mode inferior-buffer basedir))
           (julia-repl--run-hooks inferior-buffer)
-          (setf (buffer-local-value 'julia-repl--inferior-buffer-suffix inferior-buffer) suffix)
+          (with-current-buffer inferior-buffer
+            (setq-local julia-repl--inferior-buffer-suffix suffix))
           inferior-buffer)))))
 
 ;;;###autoload
@@ -538,7 +649,21 @@ Valid keys are the first items in ‘julia-repl-executable-records’."
 
 This is the standard entry point for using this package."
   (interactive)
-  (pop-to-buffer (julia-repl-inferior-buffer)))
+  (let ((script-buffer (current-buffer))
+	(inferior-buffer (julia-repl-inferior-buffer)))
+    (with-current-buffer inferior-buffer
+      (setq julia-repl--script-buffer script-buffer))
+    (if julia-repl-pop-to-buffer
+	(pop-to-buffer inferior-buffer)
+      (switch-to-buffer inferior-buffer))))
+
+(defun julia-repl--switch-back ()
+  "Switch to the buffer that was active before last call to `julia-repl'."
+  (interactive)
+  (when (buffer-live-p julia-repl--script-buffer)
+    (if julia-repl-pop-to-buffer
+	(switch-to-buffer-other-window julia-repl--script-buffer)
+      (switch-to-buffer julia-repl--script-buffer))))
 
 ;;
 ;; path rewrites
@@ -593,11 +718,18 @@ Unless NO-BRACKETED-PASTE, bracketed paste control sequences are used."
   (when (eq no-newline 'prefix)
     (setq no-newline current-prefix-arg))
   (let ((inferior-buffer (julia-repl-inferior-buffer)))
-    (display-buffer inferior-buffer)
+    (if julia-repl-pop-to-buffer
+	(display-buffer inferior-buffer))
     (julia-repl--send-to-backend julia-repl--terminal-backend
                                  inferior-buffer (s-trim string) (not no-bracketed-paste)
                                  (not no-newline))))
 
+(defun julia-repl--forward-skip-comments ()
+  "Move one line forward, then skip any comments when `julia-repl-skip-comments' is set."
+  (forward-line)
+  (when julia-repl-skip-comments
+    (forward-comment (buffer-size))))
+
 (defun julia-repl-send-line ()
   "Send the current line to the Julia REPL term buffer.
 
@@ -608,7 +740,7 @@ bracketed paste.  Unless you want this specifically, you should
 probably be using `julia-repl-send-region-or-line'."
   (interactive)
   (julia-repl--send-string (thing-at-point 'line t) 'prefix t)
-  (forward-line))
+  (julia-repl--forward-skip-comments))
 
 (defun julia-repl-send-region-or-line (&optional prefix suffix)
   "Send active region (if any) or current line to the inferior buffer.
@@ -628,7 +760,7 @@ and after."
           (deactivate-mark))
       (progn
         (-send-string (thing-at-point 'line t))
-        (forward-line)))))
+        (julia-repl--forward-skip-comments)))))
 
 (defun julia-repl-edit ()
   "Call @edit on the expression.
@@ -680,7 +812,7 @@ If a buffer corresponds to a file and is not saved, the function prompts the use
                 (save-buffer)
               (unless (file-exists-p file)
                 (message "need to save the file first"))))
-          (julia-repl--send-string (concat "Revise.includet(\""
+          (julia-repl--send-string (concat "using Revise; Revise.includet(\""
                                            (julia-repl--path-rewrite file julia-repl-path-rewrite-rules)
                                            "\");")))
       (message "buffer does not correspond to a file"))))
@@ -743,6 +875,14 @@ name), separated by dots, as a list."
 	(with-current-buffer (julia-repl-inferior-buffer) (cd directory)))
     (warn "buffer not associated with a file")))
 
+(cl-defun julia-repl-find-projectfile (&optional (start-filename (buffer-file-name)))
+  "Find a project file in the parent directories of the `start-filename'. When no such file is found, return `nil' "
+  (when start-filename
+    (cl-flet ((find-projectfile (project-filename)
+                (locate-dominating-file start-filename project-filename)))
+      (or (find-projectfile "Project.toml")
+          (find-projectfile "JuliaProject.toml")))))
+
 (defun julia-repl-activate-parent (arg)
   "Look for a project file in the parent directories, if found, activate the project.
 
@@ -752,16 +892,13 @@ When called with a prefix argument, activate the home project."
       (progn
         (message "activating home project")
         (julia-repl--send-string "import Pkg; Pkg.activate()"))
-    (cl-flet ((find-projectfile (filename)
-                                (locate-dominating-file (buffer-file-name) filename)))
-      (if-let ((projectfile (or (find-projectfile "Project.toml")
-                                (find-projectfile "JuliaProject.toml"))))
-          (progn
-            (message "activating %s" projectfile)
-            (julia-repl--send-string
-             (concat "import Pkg; Pkg.activate(\""
-                     (expand-file-name (file-name-directory projectfile)) "\")")))
-        (message "could not find project file")))))
+    (if-let ((projectfile (julia-repl-find-projectfile)))
+        (progn
+          (message "activating %s" projectfile)
+          (julia-repl--send-string
+           (concat "import Pkg; Pkg.activate(\""
+                   (expand-file-name (file-name-directory projectfile)) "\")")))
+      (message "could not find project file"))))
 
 (defun julia-repl-set-julia-editor (editor)
   "Set the JULIA_EDITOR environment variable."
@@ -773,10 +910,45 @@ When called with a prefix argument, activate the home project."
   (interactive)
   (julia-repl--send-string "ENV[\"JULIA_EDITOR\"] = \"emacsclient\";"))
 
+(defun julia-repl--show (kind mime base64data)
+  "Show data sent from Julia via EmacsVterm.jl in some Emacs window.
+KIND identifies the type of data being sent, MIME is the mime
+type of the data and BASE64DATA contains the actual
+Base64-encoded data.
+
+Currently, only showing the documentation is supported, but
+later, things like showing inline images or rendered LaTeX might
+be added."
+  (pcase `(,kind . ,mime)
+    ('("documentation" . "text/html")
+     (with-current-buffer-window "*julia-doc*" nil nil
+       (insert (decode-coding-string (base64-decode-string base64data) 'utf-8))
+       (shr-render-region (point-min) (point-max))
+       (goto-char (point-min))
+       (view-mode-enter)))
+    (`("image" . ,_)
+     (with-current-buffer-window "*julia-img*" nil nil
+       (let ((imgdata (base64-decode-string base64data)))
+         ;; TODO: Implement different display modes like julia-snail
+         ;; allow images to be erased
+         (fundamental-mode)
+         (read-only-mode -1)
+         (erase-buffer)
+         (insert imgdata)
+         (image-mode)
+         (image-transform-fit-to-window)
+         (add-hook 'window-size-change-functions (lambda (_) (image-transform-fit-to-window)) 0 t))))
+    (_ (error "Unsupported data kind `%s' or MIME type `%s' (upgrade julia-repl or use older EmacsVterm.jl)"
+              kind mime))))
+
 ;;;###autoload
 (define-minor-mode julia-repl-mode
-  "Minor mode for interacting with a Julia REPL running inside a term."
-  nil ">"
+  "Minor mode for interacting with a Julia REPL running inside a term.
+
+\\{julia-repl-mode-map}"
+  :init-value nil
+  :lighter " >"
+  :keymap
   `((,(kbd "C-c C-a")    . julia-repl-activate-parent)
     (,(kbd "C-c C-b")    . julia-repl-send-buffer)
     (,(kbd "C-c C-c")    . julia-repl-send-region-or-line)
